#include "gpu.h"

#include "runtime/config.h"
#include "runtime/threading.h"
#include "runtime/handles.h"

#include "gfx/renderer_api.h"

#include <stdlib.h>

RT_CVAR_I(rt_VkMaxSemaphores, "Maximum number of semaphores. Default: 1024", 1024);

typedef struct rt_gpu_semaphore_s {
    uint32_t version;
    VkSemaphore semaphore;
    uint64_t current_value;
    /* TODO: State tracking. We want to ensure that we don't introduce gpu hangs by waiting on a not-signaled semaphore. */

    struct rt_gpu_semaphore_s *next_free;
} rt_gpu_semaphore;

static rt_gpu_semaphore *_semaphores;
static rt_gpu_semaphore *_first_free;
static rt_mutex *_lock;

static void DestroySemaphore(rt_gpu_semaphore *s) {
    vkDestroySemaphore(g_gpu.device, s->semaphore, g_gpu.alloc_cb);
    s->semaphore = VK_NULL_HANDLE;
    rtLockMutex(_lock);
    s->next_free = _first_free;
    _first_free  = s;
    rtUnlockMutex(_lock);
}

rt_result InitializeSempahoreManagement(void) {
    _semaphores = calloc(rt_VkMaxSemaphores.i, sizeof(rt_gpu_semaphore));
    if (!_semaphores)
        return RT_OUT_OF_MEMORY;

    _lock = rtCreateMutex();
    if (!_lock) {
        free(_semaphores);
        return RT_UNKNOWN_ERROR;
    }

    /* Keep 0 unused for the invalid handle */
    _first_free = &_semaphores[1];
    for (int i = 1; i < rt_VkMaxSemaphores.i - 1; ++i)
        _semaphores[i].next_free = &_semaphores[i + 1];
    _semaphores[rt_VkMaxSemaphores.i - 1].next_free = NULL;

    return RT_SUCCESS;
}

void ShutdownSemaphoreManagement(void) {
    for (int i = 1; i < rt_VkMaxSemaphores.i; ++i) {
        vkDestroySemaphore(g_gpu.device, _semaphores[i].semaphore, g_gpu.alloc_cb);
    }
}

rt_result RT_RENDERER_API_FN(CreateSemaphores)(uint32_t count, const rt_gpu_semaphore_info *info, rt_gpu_semaphore_handle *p_semaphores) {
    for (uint32_t i = 0; i < count; ++i) {
        rtLockMutex(_lock);
        rt_gpu_semaphore *sem = _first_free;
        if (sem)
            _first_free = sem->next_free;
        rtUnlockMutex(_lock);

        if (!sem) {
            for (uint32_t j = 0; j < i; ++j) {
                uint32_t index = p_semaphores[j].index;
                DestroySemaphore(&_semaphores[index]);
            }
            return RT_OUT_OF_MEMORY;
        }

        sem->version = (sem->version + 1) % RT_RENDER_BACKEND_HANDLE_MAX_VERSION;
        
        VkSemaphoreTypeCreateInfo type_info = {
            .sType         = VK_STRUCTURE_TYPE_SEMAPHORE_TYPE_CREATE_INFO,
            .semaphoreType = VK_SEMAPHORE_TYPE_TIMELINE,
            .initialValue  = info[i].initial_value,
        };
        VkSemaphoreCreateInfo semaphore_info = {
            .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO,
            .pNext = &type_info,
        };
        
        if (vkCreateSemaphore(g_gpu.device, &semaphore_info, g_gpu.alloc_cb, &sem->semaphore) !=
            VK_SUCCESS) {
            for (uint32_t j = 0; j < i; ++j) {
                uint32_t index = p_semaphores[j].index;
                DestroySemaphore(&_semaphores[index]);
            }
            return RT_UNKNOWN_ERROR;
        }

        p_semaphores[i].version = sem->version;
        p_semaphores[i].index   = (uint32_t)(sem - _semaphores);
    }
    return RT_SUCCESS;
}

void RT_RENDERER_API_FN(DestroySemaphores)(uint32_t count, rt_gpu_semaphore_handle *semaphores) {
    for (uint32_t i = 0; i < count; ++i) {
        uint32_t index = semaphores[i].index;
        if (index >= (uint32_t)rt_VkMaxSemaphores.i)
            continue;
        if (semaphores[i].version != _semaphores[index].version) {
            rtLog("vk",
                  "Tried to destroy semaphore %u with version %u, but the semaphore has version %u",
                  index,
                  semaphores[i].version,
                  _semaphores[index].version);
            continue;
        }
        DestroySemaphore(&_semaphores[index]);
    }
}

VkSemaphore rtGetSemaphore(rt_gpu_semaphore_handle handle) {
    uint32_t index = handle.index;
    if (!RT_IS_HANDLE_VALID(handle) || index >= (uint32_t)rt_VkMaxSemaphores.i)
        return VK_NULL_HANDLE;
    if (_semaphores[index].version != handle.version)
        return VK_NULL_HANDLE;
    return _semaphores[index].semaphore;
}

uint64_t RT_RENDERER_API_FN(GetSemaphoreValue)(rt_gpu_semaphore_handle semaphore) {
    uint32_t index = semaphore.index;
    if (!RT_IS_HANDLE_VALID(semaphore) || index >= (uint32_t)rt_VkMaxSemaphores.i)
        return 0;
    if (_semaphores[index].version != semaphore.version)
        return 0;
    vkGetSemaphoreCounterValue(g_gpu.device,
                               _semaphores[index].semaphore,
                               &_semaphores[index].current_value);
    return _semaphores[index].current_value;
}